/*
This project provide access to non public fields
and methods as help for writing tests.
Copyright (C) 2007  Christof Lehmann

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package net.sf.testextensions;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;

/**
 * The <code>AccessMethodImpl</code> provide access to the non public methods
 * of an object.
 *
 * @author forge-cl
 *
 */
class AccessMethodImpl implements AccessMethod {
	/**
	 * The class <code>MethodContainer</code> help accessing a method.
	 *
	 * @author forge-cl
	 *
	 */
	private class MethodContainer extends AccessibleObjectContainer {
		/**
		 * The field <code>peerMethod</code> memorize the method, which
		 * can be invoked.
		 */
		private Method peerMethod;

		/**
		 * The field <code>manipulator</code> memorize the required manipulator
		 * of the SecurityManager.
		 */
		private AccessibilityManipulator manipulator;

		/**
		 * The constructor initialize the MethodContainer.
		 * @param aMethod the parameter contains the method,
		 * 	which should be invoced.
		 */
		public MethodContainer(final Method aMethod) {
			super(aMethod.getParameterTypes());
			peerMethod = aMethod;
			manipulator = AccessibilityManipulator
					.selectManipulator(peerMethod);
		}

		/**
		 * The method <code>call</code> invoke the method.
		 * @param args the arguments of the method invocation
		 * @return the result object of the method invocation
		 * @throws TestExtensionException if an error occures
		 */
		public Object call(final Object[] args) throws TestExtensionException {
			Object result = null;
			try {
				manipulator.before();
				result = peerMethod.invoke(inspectedObject, args);
			} catch (IllegalArgumentException e) {
				throw new TestExtensionException(e);
			} catch (IllegalAccessException e) {
				throw new TestExtensionException(e);
			} catch (InvocationTargetException e) {
				throw new TestExtensionException(e);
			} finally {
				manipulator.after();
			}
			return result;
		}

	}

	/**
	 * The field <code>inspectedClass</code> memorize the
	 * class of the inspectedObject.
	 */
	private Class inspectedClass;

	/**
	 * The field <code>inspectedObject</code> memorize the
	 * object whose methods should be invoked.
	 */
	private Object inspectedObject;

	/**
	 * The field <code>methodName</code> memorize the name of
	 * the method.
	 */
	private String methodName;

	/**
	 *
	 */
	private ArrayList < MethodContainer > methods =
		new ArrayList < MethodContainer >();

	/**
	 * The constructor initialize the object or throws an exception.
	 *
	 * @param aInspectedClass
	 *            the method is a member of this class.
	 * @param aInspectedObject
	 *            the object, whose method should be called.
	 * @param aMethodName
	 *            the name of the method, which should be accessed.
	 * @throws TestExtensionException
	 *             if error occures during initialization.
	 */
	public AccessMethodImpl(final Class aInspectedClass,
			final Object aInspectedObject, final String aMethodName)
			throws TestExtensionException {
		inspectedClass = aInspectedClass;
		inspectedObject = aInspectedObject;
		methodName = aMethodName;
		findMethods();
	}

	/**
	 * The method <code>findMethods</code> create an array
	 * with all methods of an object, whose name is the declared
	 * name.
	 * @throws TestExtensionException if no method was found
	 *
	 */
	private void findMethods() throws TestExtensionException {
		for (Method m : inspectedClass.getDeclaredMethods()) {
			if (m.getName().compareTo(methodName) == 0) {
				methods.add(new MethodContainer(m));
			}
		}
		if (methods.isEmpty()) {
			throw new TestExtensionException("No method found!");
		}
	}

	/**
	 * The method <code>call</code> the definied methods with the spezified
	 * arguments.
	 *
	 * @param args
	 *            the args for the method call.
	 * @return the return value of the method call
	 * @throws TestExtensionException
	 *             if no method, which suites to the args, was found or any
	 *             other error occures.
	 */
	public final Object call(final Object[] args)
		throws TestExtensionException {
		MethodContainer curMethod = null;
		for (MethodContainer m : methods) {
			if (m.suitesToArgs(args)) {
				curMethod = m;
				break;
			}
		}
		if (curMethod == null) {
			throw new TestExtensionException("");
		} else {
			return curMethod.call(args);
		}
	}

	/**
	 * The method <code>call</code> the definied methods with no arguments.
	 *
	 * @return the return value of the method call
	 * @throws TestExtensionException
	 *             if no method, which suites to the args, was found or any
	 *             other error occures.
	 */

	public final Object call() throws TestExtensionException {
		return call(new Object[] {});
	}

	/**
	 * The method <code>call</code> the definied methods with the spezified
	 * arguments.
	 *
	 * @param arg
	 *            one parameter for the method call.
	 * @return the return value of the method call
	 * @throws TestExtensionException
	 *             if no method, which suites to the args, was found or any
	 *             other error occures.
	 */
	public final Object call(final Object arg) throws TestExtensionException {
		return call(new Object[] { arg });
	}

	/**
	 * The method <code>call</code> the definied methods with the spezified
	 * arguments.
	 *
	 * @param arg1
	 *            first parameter for the method call.
	 * @param arg2
	 *            second parameter for the method call.
	 * @return the return value of the method call
	 * @throws TestExtensionException
	 *             if no method, which suites to the args, was found or any
	 *             other error occures.
	 */
	public final Object call(final Object arg1, final Object arg2)
			throws TestExtensionException {
		return call(new Object[] { arg1, arg2 });
	}

	/**
	 * The method <code>call</code> the definied methods with the spezified
	 * arguments.
	 *
	 * @param arg1
	 *            first parameter for the method call.
	 * @param arg2
	 *            second parameter for the method call.
	 * @param arg3
	 *            third parameter for the method call.
	 *
	 * @return the return value of the method call
	 * @throws TestExtensionException
	 *             if no method, which suites to the args, was found or any
	 *             other error occures.
	 */
	public final Object call(final Object arg1, final Object arg2,
			final Object arg3) throws TestExtensionException {
		return call(new Object[] { arg1, arg2, arg3 });
	}

}
