/*
 This project provide access to non public fields
 and methods as help for writing tests.
 Copyright (C) 2007  Christof Lehmann

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package net.sf.testextensions;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

/**
 * The class <code>AccessObjectImp</code> help accessing non public members of
 * objects.
 *
 * @author forge-cl
 *
 */
class AccessObjectImpl implements AccessObject {

	/**
	 * The field <code>inspectedClass</code> remember the class of the
	 * instpected object.
	 */
	private Class inspectedClass;

	/**
	 * The field <code>inspectedObject</code> remember the inspected object.
	 */
	private Object inspectedObject;

	/**
	 * The field <code>constructor</code> remember the selected constructor
	 * for initiating the object.
	 */
	private Constructor constructor;

	/**
	 * The constructor initiate a new instance of the AccessObjectImpl.
	 *
	 * @param aClassName
	 *            declares the class name of the new object
	 * @param theArguments
	 *            definites the arguments of the constructor.
	 * @throws TestExtensionException
	 *             shows, that anything goes wrong.
	 */
	AccessObjectImpl(final String aClassName, final Object[] theArguments)
			throws TestExtensionException {
		try {
			inspectedClass = Class.forName(aClassName);
			selectConstructor(theArguments);
			if (constructor == null && theArguments.length == 0) {
				inspectedClass.newInstance();
			} else if (constructor == null) {
				throw new TestExtensionException("");
			}
			initObject(theArguments);
		} catch (SecurityException e) {
			throw new TestExtensionException(e);
		} catch (IllegalArgumentException e) {
			throw new TestExtensionException(e);
		} catch (InstantiationException e) {
			throw new TestExtensionException(e);
		} catch (IllegalAccessException e) {
			throw new TestExtensionException(e);
		} catch (ClassNotFoundException e) {
			throw new TestExtensionException(e);
		}
	}

	/**
	 * The method <code>suiteParameter2ArgumentObjects</code> compares the
	 * classes of theConstructorTypes with the classes of theArgumentTypes.
	 *
	 * @param theConstructorTypes
	 *            array with the classes of the constructor
	 * @param theArgumentTypes
	 *            array with the objects as possible arguments for the
	 *            constructor
	 * @return true if every class of theConstructorTypes suites to the class of
	 *         theArgumentTypes
	 */
	private boolean suiteClassParameter2ArgumentObjects(
			final Class[] theConstructorTypes,
			final Object[] theArgumentTypes) {
		return new AccessibleObjectContainer(theConstructorTypes)
				.suitesToArgs(theArgumentTypes);
	}

	/**
	 * The method <code>initObject</code> call the constructor and watch the
	 * accessibility.
	 *
	 * @param theArguments
	 *            for the constructor call.
	 * @throws TestExtensionException
	 *             shows any reflection api error
	 */
	private void initObject(final Object[] theArguments)
			throws TestExtensionException {
		if (constructor != null) {
			AccessibilityManipulator manipulator = AccessibilityManipulator
					.selectManipulator(constructor);
			try {
				manipulator.before();
				inspectedObject = constructor.newInstance(theArguments);
			} catch (IllegalArgumentException e) {
				throw new TestExtensionException(e);
			} catch (InstantiationException e) {
				throw new TestExtensionException(e);
			} catch (IllegalAccessException e) {
				throw new TestExtensionException(e);
			} catch (InvocationTargetException e) {
				throw new TestExtensionException(e);
			} finally {
				manipulator.after();
			}
		}
	}

	/**
	 * The method <code>selectConstructor</code> search for the first
	 * constructor, which suites to the classes of the arguments.
	 *
	 * @param theArguments
	 *            the arguments for the constructor call.
	 */
	private void selectConstructor(final Object[] theArguments) {
		constructor = null;
		Constructor[] constructors = inspectedClass.getDeclaredConstructors();
		for (Constructor tmp : constructors) {
			if (suiteClassParameter2ArgumentObjects(tmp.getParameterTypes(),
					theArguments)) {
				constructor = tmp;
				break;
			}
		}
	}

	/**
	 * @param aFieldName
	 *            the name of the field, which should be selected.
	 * @throws TestExtensionException
	 *             if anything goes wrong.
	 * @return a new object for accessing a Field.
	 */
	public final AccessField getField(final String aFieldName)
			throws TestExtensionException {

		try {
			return new AccessFieldImpl(inspectedClass, inspectedObject,
					inspectedClass.getDeclaredField(aFieldName));
		} catch (SecurityException e) {
			throw new TestExtensionException(e);
		} catch (NoSuchFieldException e) {
			throw new TestExtensionException(e);
		}

	}

	/**
	 * The method <code>getMethod</code> initiate an object, which provides
	 * access to the called Method.
	 *
	 * @param methodName
	 *            the name of the Method.
	 * @return a valid AccessMethod objects
	 * @throws TestExtensionException
	 *             if no method with this name exists or any other problems
	 *             occures.
	 */
	public final AccessMethod getMethod(final String methodName)
			throws TestExtensionException {
		return new AccessMethodImpl(inspectedClass,
				inspectedObject, methodName);
	}

	/**
	 * The method <code>newInnerInstance</code> create a instance of an inner
	 * class from the current object.
	 *
	 * @param innerClassname
	 *            The name of the inner class without the prefix for the outer
	 *            class.
	 * @return a new instance of an inner class of the current object.
	 * @throws TestExtensionException
	 *             shows any error.
	 */
	public final AccessObject newInnerInstance(final String innerClassname)
			throws TestExtensionException {
		return newInnerInstance(innerClassname, new Object[] {});
	}

	/**
	 * The method <code>newInnerInstance</code> create a instance of an inner
	 * class from the current object.
	 *
	 * @param innerClassname
	 *            The name of the inner class without the prefix for the outer
	 *            class.
	 * @param args
	 *            the arguments for the constructor.
	 * @return a new instance of an inner class of the current object.
	 * @throws TestExtensionException
	 *             shows any error.
	 */
	public final AccessObject newInnerInstance(final String innerClassname,
			final Object[] args) throws TestExtensionException {
		for (Class innerClass : inspectedClass.getDeclaredClasses()) {
			if (innerClass.getName().endsWith(innerClassname)) {
				Object[] prepArgs = new Object[args.length + 1];
				prepArgs[0] = inspectedObject;
				System.arraycopy(args, 0, prepArgs, 1, args.length);
				return new AccessObjectImpl(innerClass.getName(), prepArgs);
			}
		}
		throw new TestExtensionException("inner class not found!");
	}

}
